(* ::Package:: *)

(* Pairwise Comparison (Ranking) Package for Mathematica(TM) *)
(* (C) 2014 - 2023 Konrad Kulakowski *)
(* www: www.kulakowski.org *)

(* version 0.3.5 *)

BeginPackage["PairwiseComparisons`"]

(**********************************************************************************************************)
(* Auxiliary functions *)
(**********************************************************************************************************)

ConsistentMatrixFromRank::usage = "creates consistent PC matrix from the rank list L so that m_ij = L[i]/L[j] "
RecreatePCMatrix::usage = "RecreatePCMatrix[M] recreates reciprocal matrix on the basis of upper-triagle of M";
RankToVector::usage = "TODO"
RankOrder::usage = "RankOrder[rankList] - prints order of elements on the rank list"
RankList::usage = "RankList[ranking] - prints together cardinal and ordinal result of ranking"
RankListWithPos::usage = "RankListWithPos[ranking] - prints the list of triplets: (position in the ranking, no. of the alternative, ranking value)"
RankOrderWithTies::usage = "RankOrderWithTies[rankList] - prints order of elements on the rank list. Ties are represented as sublists"
RankOrderToOrdinalRank::usage = "RankOrderToOrdinalRank[rankOrder] - converts order of elements into ordinal rank so that every alternative gets the priority which is its position in the ranking"
RankOrderToReverseOrdinalRank::usage = "RankOrderToReverseOrdinalRank[rankOrder] - converts order of elements into ordinal rank so that every alternative gets the priority which is its position in the ranking in the reverse order"
NumericalRankToOrdinalRank::usage = "NumericalRankToOrdinalRank[ranking] - converts numerical ranking into ordinal ranking in the same way as RankOrderToOrdinalRank does"
NumericalRankToReverseOrdinalRank::usage = "NumericalRankToReverseOrdinalRank[ranking] - converts numerical ranking into ordinal ranking in the same way as RankOrderToReverseOrdinalRank does"
LCMatrix::usage = "LCMatrix[size, scale] - generates so-called (by Waldek Koczkodaj) layered cake matrix - the most inconsistent matrix that could be generated in AHP"

NicePlotFraction::usage = "Plots the ratios x/y in the form of a string 'x/y' "

RescaleRanking::usage = "RescaleRanking[vect] - rescale ranking so that their components sum up to one"

SymbPCM::usage = "SymbPCM[symbol, size] - creates symbolic PC matrix. Might be useful for tracing algorithms"
SymbPCMRec::usage = "SymbPCMRec[symbol, size] - creates symbolic reciprocal PC matrix. Might be useful for tracing algorithms"

VersionPC::usage = "VersionPC[] prints current version of the package"


(**********************************************************************************************************)
(* Eigenvalue Rank Methods *)
(**********************************************************************************************************)

PrincipalEigenValue::usage = "PrincipalEigenValue[M] returns the largest principal eigenvalue of M";
PrincipalEigenValueSym::usage = "PrincipalEigenValueSym[M] is a symbolic version of PrincipalEigenValue[M]";
PrincipalEigenVector::usage = "PrincipalEigenVector[M] returns the eigenvector of M corresponding to its principal eigenvalue ";
(* 
PrincipalEigenVectorSym::usage = "PrincipalEigenVectorSym[M] is a symbolic version of PrincipalEigenVector[M]";
*)
SaatyIdx::usage = "SaatyIdx[M] returns the value of the Saaty Inconsistency Index computed for the matrix M";
SaatyIdxSym::usage = "SaatyIdxSym[M] is a symbolic version of SaatyIdx[M]";
EigenvalueRank::usage = "EigenvalueRank[M] returns the principal eigen vector of M rescaled in a way that the sum of its entries is 1";
EigenvalueRankSym::usage = "EigenvalueRankSym[M] is a symbolic version of EigenvalueRank[M]";
AHPA::usage = "AHPA[\!\(\*SubscriptBox[\(C\), \(1\)]\),...,\!\(\*SubscriptBox[\(C\), \(n\)]\)] computes rankings for one or more matrices of alternatives on the alternative levle"
AHPC::usage = "AHPC[C, AHPA[\!\(\*SubscriptBox[\(C\), \(1\)]\),...,\!\(\*SubscriptBox[\(C\), \(n\)]\)], AHPA[\!\(\*SubscriptBox[\(C\), \(1\)]\),...,\!\(\*SubscriptBox[\(C\), \(n\)]\)], AHPA[\!\(\*SubscriptBox[\(C\), \(1\)]\),...,\!\(\*SubscriptBox[\(C\), \(n\)]\)]] computes ranking for the hierarchy on the criteria level"
HPCA::usage = "AHPA[RankingFunction, \!\(\*SubscriptBox[\(C\), \(1\)]\),...,\!\(\*SubscriptBox[\(C\), \(n\)]\)] computes rankings for one or more matrices of alternatives on the alternative levle"
HPCC::usage = "AHPC[RankingFunction, C, AHPA[\!\(\*SubscriptBox[\(C\), \(1\)]\),...,\!\(\*SubscriptBox[\(C\), \(n\)]\)], AHPA[\!\(\*SubscriptBox[\(C\), \(1\)]\),...,\!\(\*SubscriptBox[\(C\), \(n\)]\)], AHPA[\!\(\*SubscriptBox[\(C\), \(1\)]\),...,\!\(\*SubscriptBox[\(C\), \(n\)]\)]] computes ranking for the hierarchy on the criteria level"

RationalRandomScaledMatrix::usage="RationalRandomScaledMatrix[matrixSize_, scaleSize_] - generates rational fully random matrix based on the rational scale of the range scaleSize"
RandomIndex::usage="RandomIndex[n_, tries_, scaleSize_] - returns montecarlo determined average inconsistency of a matrix of size n bazsed on scale given by scaleSize"
RI::usage="Same as RandomIndex[n_, tries_, scaleSize_]"
RandomIndex::usage="Returns radom index for matrix of the given size based on the fundamentala scale"
RI::usage="Same as RandomIndex[size_]"
SaatyConsistencyRatio::usage="Returns consistency ratio for the given matrix"
CR::usage="Returns consistency ratio for the given matrix"

(**********************************************************************************************************)
(* Geometric Mean Rank Methods *)
(**********************************************************************************************************)

GeometricRank::usage = "GeometricRank[M] returns rank list given as geometric means of rows of the matrix M";
GeometricRescaledRank::usage = "GeometricRescaledRank[M] returns a geometric mean rank list of M rescalled in a way that the sum of its entries is 1";
(*GCI::usage = "GCI[M, w,logbase] calculates geometric consistency index, where M - is the PC matrix, w - priority vector, logbase - the base of logarithm used for calculations"
*)

(**********************************************************************************************************)
(* HRE  (Heuristic Rating Estimation) - rankings with the reference values *)  
(**********************************************************************************************************)

HREConstantTermVector::usage = "HREConstantTermVectorEntry[M,v] returns the constant term vector for the HRE method. M is the PC Matrix, whilst v is the vector of the known alternatives. In v the unknown alternatives are marked by the value 0"
HREMatrix::usage = "A = HREMatrix[M,v] returns a matrix for the HRE method. Togeter with b=HREConstantTermVector[M, v] it forms the linear equation system Au=b where u is the partial ranking vector. M is the PC Matrix, whilst v is the vector of the known alternatives. In v the unknown alternatives are marked by the value 0"
HREPartialRank::usage = "HREPartialRank[M,v] returns values for the unknown alternatives. M is the PC Matrix, whilst v is the vector of the known alternatives. In v the unknown alternatives are marked by the value 0"
HREFullRank::usage = "HREFullRank[M,v] returns values for both: known and unknown alternatives. M is the PC Matrix, whilst v is the vector of the known alternatives. In v the unknown alternatives are marked by the value 0"
HRERescaledRank::usage = "HRERescaledRank[M,v] returns the rescaled HREFullRank list, rescalled in a way that the entries sum up to 1"

HREIncompleteMatrix::usage = "HREIncompleteMatrix[matrix, vector, emptySymb] - returns HRE auxiliart matrix. matrix - possibly incomplete PCM, vector - vector of known values, emptySymb - element representing missing value"
HREIncompleteConstantTermsVectorElement::usage = "HREIncompleteConstantTermsVectorElement[matrix, vector, pos, emptySymb] - returns element of auxiliray vector. matrix - possibly incomplete PCM, vector - vector of known values, emptySymb - element representing missing value"
HREIncompleteConstantTermsVector::usage = "HREIncompleteConstantTermsVector[DM, vector, emptySymb] - retruns complete auxiliray vector. matrix - possibly incomplete PCM, vector - vector of known values, emptySymb - element representing missing value"
HREIncompletePartialRanking::usage = "HREIncompletePartialRanking[matrix, vector, emptySymb] - returns ranking for unknown alternatives. matrix - possibly incomplete PCM, vector - vector of known values, emptySymb - element representing missing value"
HREIncompleteFullRanking::usage = "HREIncompleteFullRanking[matrix, vector, emptySymb] - returns complete ranking. matrix - possibly incomplete PCM, vector - vector of known values, emptySymb - element representing missing value"
HREIncompleteRescaledRank::usage = "HREIncompleteRescaledRank[matrix, vector, emptySymb] - returns rescaled complete ranking. matrix - possibly incomplete PCM, vector - vector of known values, emptySymb - element representing missing value"

(**********************************************************************************************************)
(* HRE Geometric (Heuristic Rating Estimation using Geometric Mean) - rankings with the reference values *)  
(**********************************************************************************************************)

HREGeomConstantTermVector::usage = "HREGeomConstantTermVectorEntry[M,v] returns the constant term vector for the HRE method. M is the PC Matrix, whilst v is the vector of the known alternatives. In v the unknown alternatives are marked by the value 0"
HREGeomMatrix::usage = "A = HREGeomMatrix[M,v] returns a matrix for the HRE Geometric method. Togeter with b=HREGeomConstantTermVector[M, v] it forms the linear equation system Au=b where u is the intermediate partial ranking vector. M is the PC Matrix, whilst v is the vector of the known alternatives. In v the unknown alternatives are marked by the value 0"
HREGeomIntermediateRank::usage = "HREGeomIntermediateRank[M,v] returns values for the unknown alternatives before they are raised to the power 10. M is the PC Matrix, whilst v is the vector of the known alternatives. In v the unknown alternatives are marked by the value 0"
HREGeomPartialRank::usage = "HREGeomPartialRank[M,v] returns values for the unknown alternatives. M is the PC Matrix, whilst v is the vector of the known alternatives. In v the unknown alternatives are marked by the value 0"
HREGeomFullRank::usage = "HREGeomFullRank[M,v] returns values for both: known and unknown alternatives. M is the PC Matrix, whilst v is the vector of the known alternatives. In v the unknown alternatives are marked by the value 0"
HREGeomRescaledRank::usage = "HREGeomRescaledRank[M,v] returns the rescaled HREGeomFullRank list, rescalled in a way that the entries sum up to 1"

(**********************************************************************************************************)
(* Koczkodaj Inconsistency Methods *)
(**********************************************************************************************************)

KoczkodajTriadIdx::usage = "KoczkodajTriadIdx[{a,b,c}] returns the Koczkodaj triad inconsistency i.e. Min[Abs[1-c/ab],Abs[1-ab/c]]";
KoczkodajIdx::usage = "KoczkodajIdx[M] returns the value of the Koczkodaj inconsistency computed for the matrix M";
KoczkodajTheWorstTriad::usage = "KoczkodajTheWorstTriad[M] returns the worst triad in M according to Koczkodaj inconsistency criterion";
KoczkodajTheWorstTriads::usage = "KoczkodajTheWorstTriads[M, n] returns n worst triads in M according to Koczkodaj inconsistency criterion";
KoczkodajConsistentTriad::usage = "KoczkodajConsistentTriad[{a,b,c}] returns the closest consistent triad";
KoczkodajImproveMatrixStep::usage = "KoczkodajImproveMatrixStep[M] returns improved (more consistent) matrix in which the most inconsistent triad is replaced by the closest consistent one";
KoczkodajIdxRecip::usage = "KoczkodajIdxRecip[M] returns the Koczkodaj triad inconsistency for reciprocal matrix only";
KoczkodajTriadInconsistency::usage = "KoczkodajTriadInconsistency[i,j,k,M] returns the Koczkodaj triad inconsistency";
KoczkodajTheMostInconsistentTriad::usage = "KoczkodajTheMostInconsistentTriad[M] returns the worst triad (the most inconsistent) in M as i,j,k indexes";

(**********************************************************************************************************)
(* Condition Of Order Preservation by Bana e Costa and Vansnick *)
(**********************************************************************************************************)

COP1ViolationList::usage = "COP1ViolationList[M,r] returns the list of indices that violate the first Condition of Order Preservation (COP) postulate (formulated by Bana e Costa and Vansnick). Here, M is the PC Matrix, whilst r is the ranking result list."

COP1Check::usage = "COP2ViolationList[M,r] it is true when the first Condition of Order Preservation (COP) postulate (formulated by Bana e Costa and Vansnick) is violated. Here, M is the PC Matrix, whilst r is the ranking result list."

COP2ViolationList::usage = "COP2ViolationList[M,r] returns the list of indices that violate the second Condition of Order Preservation (COP) postulate (formulated by Bana e Costa and Vansnick). Here, M is the PC Matrix, whilst r is the ranking result list."

COP2Check::usage = "COP2ViolationList[M,r] it is true when the second Condition of Order Preservation (COP) postulate (formulated by Bana e Costa and Vansnick) is violated. Here, M is the PC Matrix, whilst r is the ranking result list."

OrderPreservationViolationMatrix[M_,v_]::usage ="OrderPreservationViolationMatrix[M,v] calcualtes order preservation matrix. -1 means the place of violation"
OrderPreservationIdx[M_, v_]::usage ="OrderPreservationIdx[M,v] calcualtes order preservation index. Multpiled by 100% means the percent of distinct pairs (over the diagonal) that violates COP"
IntensityPreservationViolation4DPositive[M_, w_]::usage ="IntensityPreservationViolation4DPositive[C, w] returns the 4D map with violations are postively violated i.e where for C=[c_ij] and ranking vector w holds c_ij > c_pq > 1 and w_ij <= w_pq"
IntensityPreservationViolation4DNegative[M_, w_]::usage= "IntensityPreservationViolation4DNegative[C, w] returns the 4D map with violations are postively violated i.e where for C=[c_ij] and ranking vector w holds c_ij < c_pq < 1 and w_ij >= w_pq"
IntensityPreservationViolation4D[M_,w_]::usage= "join of IntensityPreservationViolation4DPositive and IntensityPreservationViolation4DNegative"
IntensityPreservationIdx[M_, v_]:="IntensityPreservationIdx[M, w] - number of intensity violations divided by n^4"

(**********************************************************************************************************)
(* Ranking Discrepancy  *)
(**********************************************************************************************************)

ErrorMatrix::usage = "ErrorMatrix[M, mju] computes matrix containing entries in form e_ij = m_ji*mju(c_i)/mju(c_j). When the matrix is consistent every e_ij equals 1"

LocalDiscrepancyMatrix::usage = "LocalDiscrepancyMatrix[M, mju] computes matrix with entries d_ij = max{e_ij - 1, 1/e_ij - 1}"

GlobalDiscrepancy::usage = "GlobalDiscrepancy[M, mju] returns maximal entry of  LocalDiscrepancyMatrix[M, mju]"

(**********************************************************************************************************)
(* Results/Judgment Aggregation  *)
(**********************************************************************************************************)

AIJgeom::usage = "AIJgeom[v1,...,vn] computes aggregation matrix/vector whose elements are geometric means of elements from input v1,...,vn. Matrices/Vectors v1,...,vn need to be of the same size" 
AIJadd::usage = "AIJadd[v1,...,vn] computes aggregation matrix/vector whose elements are arithmetic means of elements from input v1,...,vn. Matrices/Vectors v1,...,vn need to be of the same size" 

ListProduct::usage = "ListProduct[{a,b,c,d}] transfroms to a*b*c*d etc. "
AIJgeomWithPriorities::usage = "AIJgeomWithPriorities[list_of_matrices_or_vectors, list_of_priorities] - aggregate results using geometric mean assuming specified priorities"
AIJadditiveWithPriorities::usage = "AIJadditiveWithPriorities[list_of_matrices_or_vectors, list_of_priorities] - aggregate results using arithmetic mean assuming specified priorities"

GED::usage = "GED[v1,...,vn] computes group euclidian distance for n PC matrices"
GEDMin::usage = "GEDMin[v1,...,vn] computes ranking as the result of minimization of GED - group euclidian distance"

CompatibilityIndex::usage = "CompatibilityIndex[a, b] - calcualtes compatibility index for a group of matrices"

(**********************************************************************************************************)
(* Incomplete pairwise comparisons matrices  *)
(**********************************************************************************************************)

HarkerMatrix::usage = "HarkerMatrix[M] computes Harker matrix A for an incomplete PC matrix M. Then, to obtain the ranking, A is processed as a standard PC matrix using eigenvalue based method."
IrreducibleMatrixQ::usage = "IrreducibleMatrixQ[M] checks whether matrix is irreducible i.e. whether graph is strongly connected. This is an indispensible condition for ranking existence for incomplete PC matrix"
EVM::usage = "EVM[M] - this calculates eigenvalue ranking based on Harker method. When matrix is complete it is just eigenvalue method."
GMM::usage = "GMM[M] - this calculates geometric mean method based ranking for complete and incomplete matrices. Obtained rankings are optimal in sense of LLS method"
NGMM::usage = "NGMM[M] - this is a shorthand from N@GMM[M]"
LLSM::usage = "LLSM[M] - this calculates logarithmic least square method for complete and incomplete matrices"

(**********************************************************************************************************)
(* Comparing ranks *)
(**********************************************************************************************************)

KendallTauDistance::usage ="KendallTauDistance[list1, list2] computes Kendall (Bubble sort) distance between two rank lists so that e.g. list1 = {a,b,c} means that a is at the first position, b is the second and c is at the third position"
NormalizedKendallTauDistance::usage = "Computes Kendall (Bubble sort) distance divided by the number of comparisons i.e. KendallTauDistance[] / n(n-1)/2 "

SpearmanDistance::usage = "SpearmanDistance[list1, list2] computes Spearman\[CloseCurlyQuote]s footrule metric is the L1 distance between two permutations. Max value is n^2/2 "
NormalizedSpearmanDistance::usage = "NormalizedSpearmanDistance[list1, list2] computes normalized Spearman distance between two rankings"
(**********************************************************************************************************)
(* Ordinal inconsistency *)
(**********************************************************************************************************)

MaxInconsistentTriadsNoTies::usage = "MaxInconsistentTriadsNoTies[n] - returns maximal number of inconsistent aka circular triads in a tournament G=(V,E) with the size |V|=n "
MaxInconsistentTriadsWithTies::usage = "MaxInconsistentTriadsWithTies[n] - returns maximal number of inconsistent aka circular triads in a generalized tournament G=(V,Ed, Eu) with the size |V|=n, where Ed - set of directed edges, Eu - set of undirected edges"
RecreateOrdinalPCMatrix::usage = "RecreateOrdinalPCMatrix[M] - creates reciprocal ordinal PC matrix from non-reciprocal ordinal one"
CircularTriadsList::usage = "CircularTriadsList[M, cond] - returns the list of circular/non circular triads, cond might be true or false"
CircularTriadsNo::usage = "CircularTriadsNo[M] - return the number of circular triads in the given matrix M"
CardinalToOrdinalPCMatrix::usage = "CardinalToOrdinalPCMatrix[M] - converts cardinal PC matrix M to the ordinal PC matrix"
KendallIndex::usage = "KendallIndex[M] - returns the value of Kendall index. M - is an ordinal PC matrix i.e. it contains -1 or 1"
GeneralizedKendallIndex::usage="GeneralizedKendallIndex[M] - returns the value of extended Kendall index. M - is an ordinal PC matrix with ties i.e. it may contain -1, 0 or 1"

(**********************************************************************************************************)
(* Monte Carlo PC matrices *)
(**********************************************************************************************************)

RandomBoolean::usage = "returns TRUE or FALSE randomly"
RandomRankingPattern::usage = "RandomRankingPattern[size,scaleSize] - returns real number PC ranking of the given size where entries are evenly distributed with respect to reciprocity in [1/scaleSize, scaleSize]]"
RandomMatrixEntry::usage = "RandomMatrixEntry[i,j,pattern,maxdisturbance] - draws i,j random entry based on the ranking pattern with the given disturbance level "
RandomMatrix::usage = "RandomMatrix[n, pattern, dist] returns random matrix of size n by n with the upper limit of the scale s and the level of disturbance dist" 
RandomMatrix::usage = "RandomMatrix[n, dist] - returns random matrix of size n by n with the upper limit of the scale 9 and the level of disturbance dist"
RationalScale::usage = "RationalScale[s] - returns rational scale from 1/s to s"
FundamentalScale::usage = "FundamentalScale[] - returns rational scale (so-called fundamental scale) from 1/9 to 9"
RandomRationalUniformRankingPattern::usage = "RandomRationalUniformRankingPattern[n, s] - returns rational number ranking of size n over the scale from 1/s to s"
(* By design this should be (scale size)^2 otherwise matrix never be consistent, even for disturbance factor 1 *)
ListOfPossibleRationalEntries::usage = "ListOfPossibleRationalEntries[entriesLimit] - returns the list of possible rational entries in rational PC matrix. Note that for the ranking over the scale 1/s to s the entriesLimit should be set to s^2"
RandomRationalMatrixEntry::usage = "RandomRationalMatrixEntry[i,j,pattern, scaleSize, maxdisturbance] - returns i,j random rational matrix entry over the given ranking pattern, scale and the level of disturbance"
RandomRationalMatrix::usgae = "RandomRationalMatrix[n, s, pattern, dist] - returns random rational matrix of the given size n by n, over the scale 1/s to s, for the given ranking pattern and the disturbance level not smaller than 1"
RandomRationalMatrix::usgae = "RandomRationalMatrix[n, dist] - returns random rational matrix of the given size n by n, over the scale 1/9 to 9, for the rational ranking patter over the scale 1/9 to 9 and the disturbance level dist not smaller than 1"
RandomlyDecompletedMatrix::usage = "RandomlyDecompletedMatrix[M, x, emptySymb] - returns randomly decompleted matrix M with x missing comparisons and missing comparisons represented by emptySymb"

(**********************************************************************************************************)
(* Different Inconsistency indices for complete and incomplete PC matrices *)
(**********************************************************************************************************)

GKI::usage = "GKI[M] - computes generalized Koczkodaj inconsistency index for incomplete PC matrices"
GeneralizedKoczkodajIndex::usage - "GeneralizedKoczkodajIndex[M] = GKI[M] - computes generalized Koczkodaj inconsistency index for incomplete PC matrices"

GCI::usage = "GCI[M] - calculates geometric consistency index for matrix M"
GeometricConsistencyIndex::usage = "GeometricConsistencyIndex[M] = GCI[M] - calculates geometric consistency index for matrix M"
GGCI1::usage = "GGCI1[M] - calculates generalized geometric consistency index for (incomplete) matrix M"
GGCI2::usage = "GGCI2[M] - calculates generalized geometric consistency index for (incomplete) matrix M"
GeneralizedGeometricConsistencyIndex1::usage = "GGCI1[M] - calculates generalized geometric consistency index for (incomplete) matrix M"
GeneralizedGeometricConsistencyIndex2::usage = "GGCI2[M] - calculates generalized geometric consistency index for (incomplete) matrix M"

HCI::usage = "HCI[M] - calculates harmonic cinsistency index for M"
HarmonicConsistencyIndex::usage = "HarmonicConsistencyIndex[M] = HCI[M] - calculates harmonic cinsistency index for M"

CIStar::usage = "CIStar[M] - calculates index of determinants for M"
IndexOfDeterminants::usage = "IndexOfDeterminants[M] = CIStar[M] - calculates index of determinants for M"

TBII1::usage = "TBII1[M] - calculates triad based inconsistency index for M"
TriadBasedInconsistencyIndex::usage = "TriadBasedInconsistencyIndex1[M] = TBII1[M] - calculates triad based inconsistency index for M"

GTBII1::usage = "GTBII1[M] - calculates generalized tuple based inconsistency index for (incomplete) M"
GeneralizedTupleBasedInconsistencyIndex1::usage = "GeneralizedTupleBasedInconsistencyIndex1[M] - calculates generalized tuple based inconsistency index for (incomplete) M"

TBII2::usage = "TBII2[M] - calculates triad based inconsistency index v. 2 for M"
TriadBasedInconsistencyIndex2::usage = "TriadBasedInconsistencyIndex2[M] = TBII2[M] - calculates triad based inconsistency index v. 2 for M"

GTBII2::usage = "GTBII1[M] - calculates generalized tuple based inconsistency index v. 2 for (incomplete) M"
GeneralizedTupleBasedInconsistencyIndex2::usage = "GeneralizedTupleBasedInconsistencyIndex2[M] = GTBII2[M] - calculates generalized tuple based inconsistency index v. 2 for (incomplete) M"

TBIIAlpha::usage = "TBIIAlpha[M, alpha] - calculates triad based inconsistency alpha index for M, 0 < alpha <1" 
TriadBasedInconsistencyIndexAlpha::usage = "TriadBasedInconsistencyIndexAlpha[matrix, alpha] = TBIIAlpha[matrix, alpha] - calculates triad based inconsistency alpha index for M, 0 < alpha <1" 

GTBIIAlpha::usage = "GTBIIAlpha[M, alpha] - calculates generalized tuple based inconsistency alpha index for (incomplete) M, 0 < alpha <1" 
GeneralizedTriadBasedInconsistencyIndexAlpha::usage = "GeneralizedTriadBasedInconsistencyIndexAlpha[matrix, alpha] = GTBIIAlpha[matrix, alpha] - calculates generalized tuple based inconsistency alpha index for (incomplete) M, 0 < alpha <1" 

GTBIIAlphaBeta::usage = "GTBIIAlphaBeta[M, alpha, beta] - calculates generalized tuple based inconsistency alpha-beta index for (incomplete) M, 0 < alpha + beta <1"
GeneralizedTriadBasedInconsistencyIndexAlphaBeta::usage - "GeneralizedTriadBasedInconsistencyIndexAlphaBeta[matrix, alpha, beta] = GTBIIAlphaBeta[matrix, alpha, beta] - calculates generalized tuple based inconsistency alpha-beta index for (incomplete) M, 0 < alpha + beta <1"

GWI::usage = "GWI[M] - calculates Golden-Wang index for M"
GoldenWangIndex::usage = "GoldenWangIndex[M] = GWI[M] - calculates Golden-Wang index for M"

GGWI::usage = "GGWI[M] - calculates generalized Golden-Wang index for (incomplete) M"
GeneralizedGoldenWangIndex::usage = "GeneralizedGoldenWangIndex[M] = GGWI[M] - calculates generalized Golden-Wang index for (incomplete) M"

SHI::usage = "SHI[M] - computes Salo-Hamalainen consistency index for M"
SaloHamalainenIndex::usage = "SaloHamalainenIndex[M] = SHI[M] - computes Salo-Hamalainen consistency index for M"

GSHI::usage = "SHI[M] - computes generlaized Salo-Hamalainen consistency index for (incomplete) M"
GeneralizedSaloHamalainenIndex::usage = "GeneralizedSaloHamalainenIndex[M] = GSHI[M] - computes generalized Salo-Hamalainen consistency index for (incomplete) M"

RE::usage = "RE[M] - calculates relative error for multiplicative matrix M" 
RelativeError::usage = "RelativeError[M] = RE[M] - calculates relative error for multiplicative matrix M"

GRE::usage = "GRE[M] - calculates generalized relative error for multiplicative (possibly incomplete) matrix M"
GeneralizedRelativeError::usage = "GeneralizedRelativeError[M] = GRE[M] - calculates relative error for multiplicative (possibly incomplete) matrix M"

RC::usage = "RC[M] - calculates consistent part of matrix M"

REv2::usage = "REv2[M] - calculates relative error v. 2 for multiplicative matrix M"
RelativeErrorv2::usage = "RelativeErrorv2[M] = REv2[M]  - calculates relative error v. 2 for multiplicative matrix M"

GREv2::usage = "GREv2[M] - calculates generalized relative error v. 2 for multiplicative (possibly incomplete) matrix M"
GeneralizedRelativeErrorv2::usage = "GeneralizedRelativeErrorv2[M] = GREv2[M] - calculates generalized relative error for multiplicative (incomplete) matrix M"


(**********************************************************************************************************)
(* PC Matrices and graph representation *)
(**********************************************************************************************************)

UpperTriangularizePCM::usage = "UpperTriangularizePCM[M] - returns PC matrix M=Mij where values for which i < j are set to infinity (not a number)"
LowerTriangularizePCM::usage = "LowerTriangularizePCM[M] - returns PC matrix M=Mij where values for which i > j are set to infinity (not a number)"
GraphOfPCM::usage = "GraphOfPCM[M] - returns weighted directed graph of the given PCM. Weights of the edge between i-th and j-th vertices  correspond to the direct comparison of i-th nad j-th alternative"
GraphOfGeneralizedPCM::usage = "GraphOfGeneralizedPCM[M] - returns directed graph of the given PCM, which corresponds to the map of comparisons. The type of labels is not important"
PlotPCGraph::usage = "PlotPCGraph[M, direction, layout, edge label formatter, vertex label formatter] or PlotPCGraph[M] - plots graph of the given PCM. Direction True/False indicates whether the graphs should be directed or not. Layout means the layout of a graph. When called PlotPCGraph[M] - plots graph of the given PCM. Direction is False, layout is Automatic"
FindAllSpanningTrees::usage = "FindAllSpanningTrees[graph] - returns all spanning trees of the given graph. Inefficient for larger graphs."
PlotUndirectedWeightedGraph::usage = "PlotUndirectedWeightedGraph[graph, matrix, layout, edgeLabelFormatter, vertexLabelFormatter] or PlotUndirectedWeightedGraph[graph, matrix]- graph is weighted undirected graph, layout - graph layout. When called as PlotUndirectedWeightedGraph[graph] layout is set to Automatic"
RankBySpanningTree::usage = "RankBySpanningTree[sptree, M] - calculates ranking induced by sptree - the spanning tree of the PC graph of the PC matrix M"
RankAndPathBySpanningTree::usage = "RankAndPathBySpanningTree[sptree, M] - calculates ranking induced by sptree - the spanning tree of the PC graph of the PC matrix M"
GetRandomSpanningTreeFromPCMatrix::usage = "GetRandomSpanningTreeFromPCMatrix[M] - returns a random spanning tree obtained from a graph of PC matrix M"


(* GetMatrixEntry::usage = "GetMatrixEntry[i ,j, M] - the same as M[i,j]"
DeleteRows::usage = "TODO"
DeleteColumns::usage = "TODO"
DeleteRowsAndColumns::usage = "TODO"
SetDiagonal::usage = "TODO"
*)


Begin["`Private`"]

(***************** Auxiliary function *****************)
ConsistentMatrixFromRank[rankList_]:=Array[rankList[[#1]]/rankList[[#2]] &,{Length[rankList],Length[rankList]}, {1,1}]
(*
  KendallTauDistanceForAPair[list1_, list2_,element1_, element2_]:= With[{posInList1Elem1 = First@First@Position[list1, element1], posInList1Elem2 = First@First@Position[list1, element2], posInList2Elem1 = First@First@Position[list2, element1], posInList2Elem2 = First@First@Position[list2, element2]}, If[Sign[posInList1Elem1 - posInList1Elem2] == Sign[posInList2Elem1 - posInList2Elem2], 0, 1]]
  KendallTauDistance[list1_, list2_]:=Plus @@ With[{pairsToCheck = Subsets[list1, {2}]}, Map[KendallTauDistanceForAPair[list1, list2, #[[1]], #[[2]]] & , pairsToCheck]] 
*)
(* Should work a little bit faster *)
KendallTauDistance[list1_, list2_]:= Plus@@Map[If[Sign[list1[[First@#]] - list1[[Last@#]]] != Sign[list2[[First@#]] - list2[[Last@#]]], 1, 0 ] & , Subsets[Range[Length@list1], {2}]];
NormalizedKendallTauDistance[list1_, list2_]:= With[{n = Length[list1]}, KendallTauDistance[list1, list2]/(n (n-1)/2)]
SpearmanDistance[r1_,r2_]:=Plus@@Map[Abs[First@Flatten@Position[r1, #] - First@Flatten@Position[r2, #]] &,Range[Length@r1]];
NormalizedSpearmanDistance[r1_,r2_]:=(SpearmanDistance[r1, r2]*2)/(Length@r1);

(* creates symbolic PC matrix *)
SymbPCM[symbol_, size_]:=Table[If[i==j, 1, Subscript[symbol, i,j]], {i,1,size},{j,1,size}];
SymbPCMRec[symbol_, size_]:=Table[If[i==j, 1, If[i>j, 1/Subscript[symbol, j,i],Subscript[symbol, i,j]]], {i,1,size},{j,1,size}];

(* Make vector summing up to one *)
RescaleRanking[vect_]:=vect/Plus@@vect;
(* Replaces ratios in the form of x/y into strings "x/y". Natural numbers are not changed *)
NicePlotFraction[number_]:=If[IntegerQ@number || Denominator[number]==1, ToString@number,ToString@Numerator[number] <> "/" <> ToString@Denominator[number]];

(***************** Pure auxiliary function *****************)
VersionPC[]:= "0.3.4 (devel)";

(***************** Eigenvalue based rankings - Implementation *****************) 

(* PrincipalEigenValue[matrix_] :=  Max@Abs@N@Eigenvalues[N@matrix] *)

PrincipalEigenValue[matrix_] :=  First@Chop@N@Eigenvalues[Chop@N@matrix, 1]
PrincipalEigenValueSym[matrix_] :=  First@Eigenvalues[matrix, 1]

TheSameSignQ[list_]:=Abs@Apply[Plus, Sign[list]]== Length[list] 

(* PrincipalEigenVector[matrix_] := 
 With[{x = Extract[Eigenvectors[matrix], 
  First@First@
    Position[Abs@N@Eigenvalues[matrix], 
     PrincipalEigenValue[matrix]]]}, If[TheSameSignQ[x],Abs[x],x]] *)
 
PrincipalEigenVector[matrix_] := With[{x = First@Chop@N@Eigenvectors[Chop@N@matrix, 1]}, If[TheSameSignQ[x],Abs[x],x]]
(* PrincipalEigenVectorSym[matrix_] := With[{x = First@Eigenvectors[matrix, 1]}, If[TheSameSignQ[x],Abs[x],x]] *)

SaatyIdx[matrix_] := Chop@With[{n = First[Dimensions[matrix]], alpha = PrincipalEigenValue[matrix]}, (alpha - n)/(n - 1)]
SaatyIdxSym[matrix_] := With[{n = First[Dimensions[matrix]], alpha = PrincipalEigenValueSym[matrix]}, (alpha - n)/(n - 1)]

(* EigenvalueRank[matrix_] := With[{pev = PrincipalEigenVector[matrix]}, (#/Apply[Plus, pev]) & /@ pev] *)
(* EigenvalueRankSym[matrix_] := With[{pev = PrincipalEigenVectorSym[matrix]}, (#/Apply[Plus, pev]) & /@ pev] *)
EigenvalueRank[matrix_]:=With[{vect = First@Abs@Chop@Eigenvectors[N@matrix, 1]}, vect/Plus@@vect];
EigenvalueRankSym[matrix_]:=With[{vect = First@Abs@Chop@Eigenvectors[matrix, 1]}, vect/Plus@@vect];


(* AHPRankForAlternative[altNo_, criteriaRank_, alternativesRank_]:= With[{dim = Length[criteriaRank]}, Apply[Plus, Array[criteriaRank[[#]]*  (alternativesRank[[#]])[[altNo]] &,dim]]];
AHP[criteriaMatrix_, alternativeMatrices__]:=With[{altmat = List[alternativeMatrices]},With[{criteriaRank = EigenvalueRank[criteriaMatrix], alternativesRank = Map[EigenvalueRank, altmat],dimAlt = First[Dimensions[First[altmat]]]},Array[AHPRankForAlternative[#, criteriaRank, alternativesRank] &, dimAlt]]];
*)

HPCA[method_, matrices__]:=With[{args = List[matrices], res =  method/@List[matrices]},
       If[Length@args==1, First@res, res]];
HPCC[method_, arguments__]:=With[{args = List[arguments]},
If [Length@args == 2, 
      With[{criteriaMatrix = args[[1]],subrankingMatrix=args[[2]] },First@{method[criteriaMatrix]} . subrankingMatrix], 
      With[{criteriaMatrix = args[[1]],subrankingMatrix=Rest[args] },First@{method[criteriaMatrix]} . subrankingMatrix]
   ]];

AHPA[matrices__]:= HPCA[EigenvalueRank, matrices];
AHPC[arguments__]:= HPCC[EigenvalueRank, arguments];


(***************** Geometric mean rankings - Implementation *****************)

GeometricRank[matrix_] := GeometricMean[#] & /@ matrix

(* deprecated *)  
NGeometricRank[matrix_] := N[GeometricMean[#] & /@ matrix]

GeometricRescaledRank[matrix_] := With[{gr = GeometricRank[matrix]}, (#/Apply[Plus, gr]) & /@ gr ]

(* deprecated *)  
NGeometricRescaledRank[matrix_] := N[GeometricRescaledRank[matrix]]

(* Geometric mean rank *)
(*GCI[M_, w_,logbase_]:=With[{n=Length[w]}, (2/((n-1)(n-2)))*Sum[Sum[Log[logbase, (M[[i,j]]w[[j]])/w[[i]] ]^2,{j, i+1, n}],{i,1,n-1}]];
*)
 
(***************** HRE - Rankings with the reference values - Implementation *****************)

GetMatrixEntry[i_, j_, matrix_] := Extract[Extract[matrix, i], j]

auxMakeRecip[a_, b_, matrix_] :=
 If[Extract[b, 1] < Extract[b, 2], a,
  If[Extract[b, 1] == Extract[b, 2], 1, 
   1/GetMatrixEntry[Extract[b, 2], Extract[b, 1], matrix]]]
  
 RecreatePCMatrix[upperTriangleMatrix_] := 
 MapIndexed[auxMakeRecip[#1, #2, upperTriangleMatrix] &, 
  upperTriangleMatrix, {2}] 

 RecreatePCMatrix[matrix_]:= With[{upperTriangular = UpperTriangularize[matrix,1]}, upperTriangular+LowerTriangularize[(upperTriangular+Transpose[upperTriangular] + IdentityMatrix[Length[matrix]])^(-1)]] 

(*
RecreatePCMatrix[upperTriangularMatrix_] :=
MapIndexed[If[#2[[1]] < #2[[2]], #1, 
If[#2[[1]] > #2[[2]], 
If[NumberQ[ToExpression[ToString[upperTriangularMatrix[[#2[[2]],#2[[1]]]]]]] ,1/upperTriangularMatrix[[#2[[2]],#2[[1]]]],Null],1]] &, upperTriangularMatrix, {2}]
*)

zeroQ[x_] := (ToString[Head[x]] == "Integer" ||  ToString[Head[x]] == "Real") && x == 0 
zeroIndices[knownVectorList_] :=  Map[First[Last[#]] &, Select[MapIndexed[List, knownVectorList], zeroQ[First[#]] &]] 
nonZeroValues[knownVectorList_] := Select[knownVectorList, Not[zeroQ[#]] &]
nonZeroIndices[knownVectorList_] := Map[First[Last[#]] &,  Select[MapIndexed[List, knownVectorList], Not[zeroQ[First[#]]] &]]
 
HREConstantTermVectorEntry[matrix_, knownVectorList_, j_] := 
 Apply[Plus, 
  MapThread[(#1 * #2 * (1/(First[Dimensions[matrix]] - 1))) &, 
   List[nonZeroValues[knownVectorList], 
    Map[GetMatrixEntry[j, #, matrix] &, 
     nonZeroIndices[knownVectorList]]]]]

HREConstantTermVector[matrix_, knownVector_] := 
 Map[List, 
  Map[HREConstantTermVectorEntry[matrix, Flatten[knownVector], #] &, 
   zeroIndices[Flatten[knownVector]]]]

DeleteRows[matrix_, listOfRows_] :=  Delete[matrix, Map[List, listOfRows]]

DeleteColumns[matrix_, listOfColumns_] := Map[Delete[#, Map[List, listOfColumns]] &, matrix]

DeleteRowsAndColumns[matrix_, listOfIndicesToRemove_] :=  
		   DeleteColumns[DeleteRows[matrix, listOfIndicesToRemove],   listOfIndicesToRemove]

SetDiagonal[matrix_, diagVal_] := MapIndexed[If[First[#2] == Last[#2], diagVal, #1] &, matrix, {2}]
 
HREMatrix[matrix_, knownVector_] := 
 SetDiagonal[(-1/(First[Dimensions[matrix]] - 1))*
   DeleteRowsAndColumns[matrix, nonZeroIndices[Flatten[knownVector]]], 1]

HREPartialRank[matrix_, knownVector_] := 
 LinearSolve[HREMatrix[matrix, knownVector], 
  HREConstantTermVector[matrix, knownVector]]

indexedUnknownValues[hrePartialResult_, knownVector_] := 
 MapThread[
  List, {Flatten[hrePartialResult], 
   zeroIndices[Flatten[knownVector]]}]

indexedKnownValues[knownVector_] := 
 MapThread[
  List, {nonZeroValues[Flatten[knownVector]], 
   nonZeroIndices[Flatten[knownVector]]}]

indexedValues[hrePartialResult_, knownVector_] := 
 Sort[Join[indexedUnknownValues[hrePartialResult, knownVector], 
   indexedKnownValues[knownVector]], #1[[2]] < #2[[2]] &]

HREFullRank[matrix_, knownVector_] := 
 Map[First[#] &, 
  indexedValues[HREPartialRank[matrix, knownVector], knownVector]]

HRERescaledRank[matrix_, knownVector_] :=
  With[{fullRank = HREFullRank[matrix, knownVector]}, (#/Apply[Plus, fullRank]) & /@ fullRank]
  
  
(***************** Incomplete HRE - Rankings with the reference values - Implementation *****************)

DeleteRowsAlt[matrix_, listOfRows_] :=  Delete[matrix, Map[List, listOfRows]]
DeleteColumnsAlt[matrix_, listOfRows_] :=  Transpose@DeleteRowsAlt[Transpose@matrix, listOfRows];
DeleteRowsAndColumnsAlt[matrix_, listToDel_] :=  DeleteColumnsAlt[DeleteRowsAlt[matrix, listToDel], listToDel];
MissingElementsByRows[M_, emptySymb_]:=Map[Plus@@# &,Map[If[ToString[#] ==ToString[emptySymb], 1, 0]&, M, {2}]];
NonZeroPositionsInVector[v_]:=Flatten@Position[Map[If[# == 0, 1, 0]&,Flatten@Transpose@v], 0];
NonZeroValuesInVector[v_]:=Select[Flatten@Transpose@v, NumberQ@ToExpression[#] &&(ToExpression[#] != 0) &];
ZeroPositionsInVector[v_]:=Flatten@Position[Map[If[# != 0, 1, 0]&,Flatten@Transpose@v], 0];

HREIncompleteMatrix[DM_, vector_, emptySymb_]:=With[
	{minor = DeleteRowsAndColumnsAlt[DM, NonZeroPositionsInVector@vector], 
	 S = Delete[MissingElementsByRows[DM, emptySymb], 
	 List/@ NonZeroPositionsInVector@vector ]}, 
		MapIndexed[If[#2[[1]]==#2[[2]], 1, If[ToString[#1]== ToString[emptySymb], 0, -(1/(Length@DM -S[[First@#2]]-1))*#1]] &, minor, {2}]];
		
HREIncompleteConstantTermsVectorElement[DM_,vect_,pos_, emptySymb_]:=Plus@@With[
	{vector=vect, 
	 miselem = MissingElementsByRows[DM, emptySymb]},
	 Map[1/(Length@DM-miselem[[pos]]-1)*If[ToString[DM[[pos,#]]] == ToString[emptySymb], 0, DM[[pos,#]]]*Flatten[Transpose[vector]][[#]] &, NonZeroPositionsInVector[vector]]];
	 
HREIncompleteConstantTermsVector[DM_, vect_, emptySymb_]:= List/@Map[HREIncompleteConstantTermsVectorElement[DM, vect, #, emptySymb] &, ZeroPositionsInVector[vect]];

HREIncompletePartialRanking[matrix_, knownVector_, emptySymb_]:= LinearSolve[
	HREIncompleteMatrix[matrix, knownVector, emptySymb],  
	HREIncompleteConstantTermsVector[matrix, knownVector, emptySymb]];
	
indexedUnknownValuesAlt[hrePartialResult_, knownVector_] := 
 MapThread[
  List, {Flatten[hrePartialResult], 
   ZeroPositionsInVector[Flatten[knownVector]]}]
   
indexedKnownValuesAlt[knownVector_] := 
 MapThread[
  List, {NonZeroValuesInVector[Flatten[knownVector]], 
   NonZeroPositionsInVector[Flatten[knownVector]]}]

indexedValuesAlt[hrePartialResult_, knownVector_] := 
 Sort[Join[indexedUnknownValuesAlt[hrePartialResult, knownVector], 
   indexedKnownValuesAlt[knownVector]], #1[[2]] < #2[[2]] &]
   
HREIncompleteFullRanking[matrix_, knownVector_, emptySymb_] := 
 Map[First[#] &, 
  indexedValuesAlt[HREIncompletePartialRanking[matrix, knownVector, emptySymb], knownVector]]
  
HREIncompleteRescaledRank[matrix_, knownVector_, emptySymb_] :=
  With[{fullRank = HREIncompleteFullRanking[matrix, knownVector, emptySymb]}, (#/Apply[Plus, fullRank]) & /@ fullRank]

(***************** HRE Geom - Rankings with the reference values - Implementation *****************)

HREGeomMatrix[matrix_, knownVector_]:= With[{dim1 = First[Dimensions[matrix]], 
	dim2 = Length[Select[Flatten[knownVector], zeroQ[#] & ]]},
		Map[If[#==0,-1,#]&,(dim1-1)*IdentityMatrix[dim2], {2}]] 

HREGeomConstantTermVectorEntry[matrix_, knownVector_, j_]:=With[{jthRow = Drop[Extract[matrix,{j}], {j}], 
	knownVectEntr = Select[Flatten[knownVector], Not[zeroQ[#]]  & ]},
	Log10[Times @@ Join[jthRow, knownVectEntr]]]

HREGeomConstantTermVector[matrix_, knownVector_] := 
 Map[List, Map[HREGeomConstantTermVectorEntry[matrix, Flatten[knownVector], #] &, 
   zeroIndices[Flatten[knownVector]]]]

HREGeomIntermediateRank[matrix_, knownVector_]:=LinearSolve[HREGeomMatrix[matrix, knownVector],
	HREGeomConstantTermVector[matrix, knownVector]]

HREGeomPartialRank[matrix_, knownVector_]:= Map[10^# &, HREGeomIntermediateRank[matrix, knownVector]]

HREGeomFullRank[matrix_, knownVector_] := 
 Map[First[#] &, indexedValues[HREGeomPartialRank[matrix, knownVector], knownVector]]

HREGeomRescaledRank[matrix_, knownVector_] := With[{fullRank = HREGeomFullRank[matrix, knownVector]},
	Map[#/(Plus @@ fullRank) &, fullRank]]

(***************** Koczkodaj inconsistency methods - Implementation *****************)

KoczkodajTriadIndex[matrix_, triad_]:= With[{i = triad[[1]], k=triad[[2]], j=triad[[3]]}, With[{a=matrix[[i,k]], b=matrix[[k,j]], c=matrix[[i,j]]}, Min[Abs[1-(a*b)/c],Abs[1-c/(a*b)]]]];
KoczkodajIndex[matrix_]:=With[{allTriads = Subsets[Range[Length@matrix], {3}]}, With[{existingTriads = Select[allTriads,TriadExistsQ[matrix, #] & ]}, Max@Map[KoczkodajTriadIndex[matrix, #] &,  existingTriads]]];
KoczkodajIdx[matrix_] := KoczkodajIndex[matrix];

(*** Below there is a bit depricated implementation  ***)
KoczkodajTriadIdx[triad_] := 
 Min[Abs[1 - (Extract[triad, {1}]*Extract[triad, {2}])/
     Extract[triad, {3}]], 
  Abs[1 - Extract[
      triad, {3}]/(Extract[triad, {1}]*Extract[triad, {2}])]]    

auxUniqueTuples[n_] := 
 Select[Tuples[Range[n], 3], 
  Extract[#, 1] != Extract[#, 2] &&   
    Extract[#, 1] != Extract[#, 3] &&  
    Extract[#, 2] != Extract[#, 3] &] 
 
auxMakeATriad[matrix_, tuple_] := 
 List[Extract[Extract[matrix , Extract[tuple, 1]], 
   Extract[tuple, 2]], 
  Extract[Extract[matrix , Extract[tuple, 2]], Extract[tuple, 3]], 
  Extract[Extract[matrix , Extract[tuple, 1]], Extract[tuple, 3]]] 

auxUniqueTriadsAndTuples[matrix_] := 
 List[#, auxMakeATriad[matrix, #], 
    KoczkodajTriadIdx[auxMakeATriad[matrix, #]]] & /@ 
  auxUniqueTuples[First[Dimensions[matrix]]]

KoczkodajTheWorstTriad[matrix_] := 
 First[Sort[auxUniqueTriadsAndTuples[matrix], 
   Last[#1] > Last[#2] & ] ] 

KoczkodajTheWorstTriads[matrix_, n_] := 
 Take[Sort[auxUniqueTriadsAndTuples[matrix], Last[#1] > Last[#2] & ] ,
   n]
   
KoczkodajTriadInconsistency[i_,j_,k_, matrix_]:=Min[Abs[1-(matrix[[i,j]]*matrix[[j,k]])/matrix[[i,k]]],Abs[1-matrix[[i,k]]/(matrix[[i,j]]*matrix[[j,k]])]];
TriadsBySize[size_]:=TriadsBySize[size]=Subsets[Range[size],{3}];

KoczkodajIdxRecip[matrix_]:=With[{triads = TriadsBySize[Length@matrix]},Max@Map[KoczkodajTriadInconsistency[#[[1]], #[[2]], #[[3]], matrix] &, triads]]

KoczkodajTheMostInconsistentTriad[matrix_]:=First@With[{triads = TriadsBySize[Length@matrix]},triads[[Ordering[Map[TriadInconsistency[#[[1]], #[[2]], #[[3]], matrix] &, triads], -1]]]]

KoczkodajConsistentTriad[triad_] := 
 With[{q = First[triad], r = Extract[triad, {2}], s = Last[triad]},
  List[q^(2/3) r^(-1/3) s^(1/3), q^(-1/3) r^(2/3) s^(1/3), 
   q^(1/3) r^(1/3) s^(2/3)]]

KoczkodajImproveMatrixStep[matrix_] := 
 With[{worstTriad = KoczkodajTheWorstTriad[matrix]}, 
  With[{improvedTriad = KoczkodajConsistentTriad[Extract[worstTriad, {2}]], 
    mik = Drop[First[worstTriad], {3}], 
    mkj = Drop[First[worstTriad], {1}], 
    mij = Drop[First[worstTriad], {2}]}, 
   With[{mikval = First[improvedTriad], 
     mkjval = Extract[improvedTriad, {2}], 
     mijval = Last[improvedTriad]}, 
    ReplacePart[
     ReplacePart[
      ReplacePart[
       ReplacePart[
        ReplacePart[ReplacePart[matrix, Reverse[mij] -> 1/mijval], 
         mij -> mijval], Reverse[mkj] -> 1/mkjval], mkj -> mkjval], 
      Reverse[mik] -> 1/mikval], mik -> mikval]]]]

(***************** Condition of order preservation methods - Implementation *****************)

COP1ViolationList[matrix_, resultList_] := 
 Select[Flatten[
   MapIndexed[ 
    List[((#1 <= 
           1) && (Extract[resultList, First[#2]]/
            Extract[resultList, Last[#2]] <= 1)) || ((#1 >= 
           1) && (Extract[resultList, First[#2]]/
            Extract[resultList, Last[#2]] >= 1)), #2] &, matrix, {2}],
    1], First[#] == False &]

COP1Check[matrix_, resultList_] := 
 Length[COP1ViolationList[matrix, resultList]] == 0


(*COP1ViolationList[M_, ranking_]:=With[{X1 =Sign@Chop@Log@M, 
		X2 = Sign@Chop@Log@ConsistentMatrixFromRank[ranking]}, 
		Map[#[[2]] &,Select[Flatten[MapIndexed[{#1, #2} &, X1 - X2, {2}], 1], First@# > 0 &]]];
		
COP1Check[M_, ranking_]:=With[{X1 =Sign@N@Log@M, X2 = Sign@Chop@Log@ConsistentMatrixFromRank[ranking]}, 
     Length@Select[Flatten[X1-X2], # > 0 &]\[Equal]0];*)

mij[entry_, matrix_] := 
 GetMatrixEntry[First[First[entry]], Last[First[entry]], matrix]

mkl[entry_, matrix_] := 
 GetMatrixEntry[First[Last[entry]], Last[Last[entry]], matrix]

muimuj[entry_, resultVect_] := 
 Extract[resultVect, First[First[entry]]]/
  Extract[resultVect, Last[First[entry]]]

mukmul[entry_, resultVect_] := 
 Extract[resultVect, First[Last[entry]]]/
  Extract[resultVect, Last[Last[entry]]]

auxUniqePairs[baseList_] := 
 Select[Tuples[baseList, 2], First[#] != Last[#] &]

COP2ViolationList[matrix_, resultList_] := 
 Select[Map[
   List[((mij[#, matrix] <= 
          mkl[#, matrix]) && (muimuj[#, resultList] <= 
          mukmul[#, resultList])) || ((mij[#, matrix] >= 
          mkl[#, matrix]) && (muimuj[#, resultList] >= 
          mukmul[#, resultList])), #] &, 
   auxUniqePairs[auxUniqePairs[Range[First[Dimensions[matrix]]]]]], 
  First[#] == False &]

COP2Check[matrix_, resultList_] := 
 Length[COP2ViolationList[matrix, resultList]] == 0

RankToVector[list_] :=
  Transpose[List[list]]
  
OrderPreservationViolationMatrix[M_,v_]:= Map[If[#>= 0, 0, -1] &,Sign[Log[M]]*Sign[Log[Transpose[{v}] . {1/v}]], {2}];
  
OrderPreservationIdx[M_, v_]:=With[{n=Length@v},-(Plus@@Flatten@OrderPreservationViolationMatrix[M, v]/(n(n-1)))];
IntensityPreservationViolation4DPositive[M_, w_] :=With[{n=Length@M},Table[If[M[[i,j]]-M[[p,q]] > 0 && M[[p,q]]>1, 1,0],{i,1,n},{j,1,n},{p,1,n},{q,1,n}]*Table[If[w[[i]]/w[[j]]<=w[[p]]/w[[q]],-1,0],{i,1,n},{j,1,n},{p,1,n},{q,1,n}]];
IntensityPreservationViolation4DNegative[M_, w_] :=With[{n=Length@M},Table[If[M[[i,j]]-M[[p,q]] < 0 && M[[p,q]]<1, 1,0],{i,1,n},{j,1,n},{p,1,n},{q,1,n}]*Table[If[w[[i]]/w[[j]]>=w[[p]]/w[[q]],-1,0],{i,1,n},{j,1,n},{p,1,n},{q,1,n}]];
IntensityPreservationViolation4D[M_,w_]:=Join[IntensityPreservationViolation4DPositive[M,w], IntensityPreservationViolation4DNegative[M, w]];

IntensityPreservationIdx[M_, v_]:=With[{n=Length@v},-(Plus@@Flatten@IntensityPreservationViolation4D[M, v]/n^4)];

(***************** Discrepancy - Implementation *****************)

ErrorMatrix[M_, mju_] := 
  MapIndexed[1/#1 * Extract[mju, First[#2]]/Extract[mju, Last[#2]] &, M, {2}]

LocalDiscrepancyMatrix[M_, mju_] := 
  MapIndexed[Max[#1 - 1, 1/#1 - 1] & , ErrorMatrix[M, mju], {2}]

GlobalDiscrepancy[M_, mju_] := Max[Flatten[LocalDiscrepancyMatrix[M, mju]]]

RankOrder[rankList_] := Flatten[Map[Last[#] &,
	Reverse[SortBy[MapIndexed[{#1, #2} &, rankList],First[#1] &]]]]
	
RankList[ranking_]:=Reverse@SortBy[Transpose@{Range[Length@ranking],ranking}, Last];

RankListWithPos[ranking_]:=Map[Flatten@# &, Transpose@{Range[Length@ranking],Reverse@SortBy[Transpose@{Range[Length@ranking],ranking}, Last]}];

(*  Auxiliary function for RankOrderWithTies *)
CleanElemens[tmpl_]:={First@First@tmpl, Map[First@#[[2]] &, tmpl]};

RankOrderWithTies[rankList_]:=Map[If[Length@#[[2]]==1,First@#[[2]], #[[2]]] &,
	Reverse@SortBy[Map[CleanElemens@# & , Values@GroupBy[MapIndexed[{#1, #2} &, rankList], First]], First]];

RankOrderToOrdinalRank[rankOrder_]:=  With[{res1 = Map[{If[ListQ@#[[1]], #[[1]], {#[[1]]}], First@#[[2]]} &,MapIndexed[{#1, #2} &, rankOrder]]},
	With[{res2 = SortBy[res1, First@First@# &]},
		Flatten@Map[Table[#[[2]], Length@#[[1]]] &,res2]]];

RankOrderToReverseOrdinalRank[rankOrder_]:=  RankOrderToOrdinalRank[Reverse@rankOrder];
NumericalRankToOrdinalRank[numRank_]:= RankOrderToOrdinalRank@RankOrderWithTies@numRank;
NumericalRankToReverseOrdinalRank[numRank_]:= RankOrderToReverseOrdinalRank@RankOrderWithTies@numRank;

LCMatrix[size_, scale_]:= Table[If[Mod[i,2]==0, If[Mod[j,2]==1, scale,1], If[Mod[j,2]==1, 1,1/scale]],{i,1,size}, {j,1,size}];

(***************** Aggregation - Implementation *****************)

(* calling with arguments separated by comas *)
AIJgeom[v__]:= Map[Power[#,1/Length[List[v]]] &, Apply[Times, List[v]]]
(* calling with many arguments given explicitly as list *)
AIJgeom[v_]:= If[ListQ@v, Apply[AIJgeom, v]];
AIJadd[v__]:= Map[# *1/Length[List[v]] &, Apply[Plus, List[v]]]

CompatibilityIndex[a_,b_]:= Mean@Flatten[a*b^-1];

ListProduct[list_]:=Product[list[[i]],{i, 1, Length@list}];
AIJgeomWithPriorities[vectors_, priorities_]:=ListProduct@Map[#[[1]]^#[[2]] &, Transpose@{vectors, priorities/Plus@@priorities}];

AIJadditiveWithPriorities[vectors_, priorities_]:=Table[priorities[[i]]/Plus@@priorities ,{i,1,Length@priorities}] . vectors;

(* Group Euclidian Distance Minimization *)
FM[pcm_]:=With[{n=Length@pcm},Sum[Sum[(pcm[[i,j]]-Subscript[w, i]/Subscript[w, j])^2,{i,1,n}],{j,1,n}]];
GED[pcms__]:=Sqrt[Plus@@FM/@{pcms}];
GEDMin[pcms__]:=Map[Last@# &,
                    Last@With[{n=Length@First@{pcms}},
                          NMinimize[{GED@pcms, 
                               Sum[Subscript[w, i],{i,1,n}]==1 && 
                               And@@Table[Subscript[w, i]>0,{i,1,n}]},
                               Table[Subscript[w, i],{i,1,n}],
                               Method->"DifferentialEvolution", AccuracyGoal->20]]];

(***************** Incomplete pairwise comparisons matrices *****************)

(* Harker method as proposed in Harker, ALTERNATIVE MODES OF QUESTIONING IN THE ANALYTIC HIERARCHY PROCESS, 
   Math. Modeling, Vol. 9. 1987 *)

HarkerMatrixPlaceHolderCount[matrix_, row_]:= Length[Select[matrix[[row]], Not[And[NumberQ[#], # >0]] &]]

HarkerMatrix[matrix_]:=MapIndexed[If[#2[[1]] == #2[[2]], HarkerMatrixPlaceHolderCount[matrix, #2[[1]]], 
If[And[NumberQ[#1], #1 >0] ,#1,0]] & , matrix, {2}] + IdentityMatrix[First[Dimensions[matrix]]]

(* Harker's method generalization *)
EVM[M_]:=EigenvalueRank[HarkerMatrix@M]; 

IrreducibleMatrixQ[M_]:=Length@ConnectedComponents@UndirectedGraph@AdjacencyGraph@Map[If[NumberQ[#],1,0] &,M,{2}]==1;

(*  
For details of the following functions see: Bozoki et al. On optimal completion of incomplete pairwise comparison matrices, Mathematical and Computer Modelling, 2010
*)

(* creates logarithmic constat term vecotr *)
IGMMConstantTermVector[M_]:=  Log[Times@@@Map[If[NumberQ[#], #, 1] &,M, {2}]];
(* creates auxiliary logarithmic matrix *)
BFR[M_]:= Table[If[i==j,1+Length@M,1],{i,1,Length@M},{j,1,Length@M}]-MapIndexed[If[ #2[[1]] == #2[[2]],#1, If[#1==0,0,1]] &, HarkerMatrix[M], {2}];
(* it solves logarithmic linear system then compute the GMM solution. It works for complete and incomplete matrices *)
GMM[M_]:=With[{roughRanking = Exp[LinearSolve[BFR[M],IGMMConstantTermVector[M]]]}, roughRanking / Plus@@roughRanking];
(* just shorthand *)
NGMM[M_]:= N@GMM[M];


(* 
LLSM implementation for incomplete PCM, according to 
Bozoki & Tsyganok, The (logarithmic) least squares optimality of the arithmetic (geometric) mean of weight vectors calculated from all spanning trees for incomplete additive (multiplicative) pairwise comparison matrices 
*)
GraphOfPCM[matrix_]:=AdjacencyGraph[Map[If[NumberQ@#,1,0] &,matrix,{2}]-IdentityMatrix[Length@matrix]];
YPart[Q_]:=Normal@KirchhoffMatrix@GraphOfPCM[Q] . Table[{Subscript[y, i]},{i,1,Length@Q}];
ComputeRowTMP[row_]:=Plus@@Log/@Select[row, NumberQ@#&]; 
BPart[Q_]:=Map[ComputeRowTMP@# &, Q];
LLSM[Q_]:=With[{llsmout = N@FindInstance[And@@MapThread[Equal[First@#1, #2] &, {YPart@Q, BPart@Q}]&&Subscript[y, 1]== 0, Table[Subscript[y, i],{i,1,Length@Q}], Reals]}, With[{unscaledRanking = Map[Exp@#[[2]] &,First@llsmout]},unscaledRanking / Plus@@unscaledRanking]];


(***************** Ordinal inconsistency *****************)

CircularTriadQ[M_, triadIdxs_]:= With[{i=triadIdxs[[1]], k=triadIdxs[[2]], j=triadIdxs[[3]]},
With[{aik=M[[i,k]], akj=M[[k,j]], aji=M[[j,i]]},With[{logList ={aik, akj, aji}},
Max[logList] - Min[logList] <=  1 && logList != {0,0,0}
]]];
MaxInconsistentTriadsNoTies[n_]:= If[OddQ[n],(n^3-n)/24,(n^3-4n)/24];
MaxInconsistentTriadsWithTies[n_]:=Binomial[n,3]-(Binomial[Floor[n/2],3] - MaxInconsistentTriadsNoTies[Floor[n/2]]) -(Binomial[Ceiling[n/2],3]  - MaxInconsistentTriadsNoTies[Ceiling[n/2]]); 
RecreateOrdinalPCMatrix[M_]:= MapIndexed[If[#2[[1]]>#2[[2]], -1*M[[ #2[[2]], #2[[1]] ]], #1 ] &, M, {2}]; 

CircularTriadsNo[M_]:= With[{dim = First@Dimensions[M]}, 
With[{triads = Subsets[Range[dim], {3}]},
Apply[Plus, Map[ If[CircularTriadQ[M, #], 1, 0] &,triads]]]];

CircularTriadsList[M_, cond_]:= With[{dim = First@Dimensions[M]}, 
With[{triads = Subsets[Range[dim], {3}]},
Select[triads, Xor[Not@cond, CircularTriadQ[M, #]] &]]];

KendallIndex[M_]:=1-CircularTriadsNo[M]/MaxInconsistentTriadsNoTies[Length@M];
GeneralizedKendallIndex[M_]:=1-CircularTriadsNo[M]/MaxInconsistentTriadsWithTies[Length@M];

CardinalToOrdinalPCMatrix[M_]:=Map[If[#==1,0,If[#>1,1,-1]] &,M, {2}];

(***************** Monte Carlo PC matrices *****************)

RandomBoolean[]:=OddQ@RandomInteger[{1,100000}];
RandomRankingPattern[size_,scaleSize_]:=Table[ If[RandomBoolean[],RandomReal[{1/scaleSize,0.99999999}],RandomReal[{1,scaleSize}] ], {i, 1,size}];
RandomMatrixEntry[i_,j_,pattern_,maxdisturbance_]:=(pattern[[i]]/pattern[[j]])*
     If[RandomBoolean[],RandomReal[{1/maxdisturbance,0.9999999}], RandomReal[{1,maxdisturbance}]];
RandomMatrix[size_, pattern_, maxdisturbance_]:=RecreatePCMatrix@Table[If[i>j,0,If[i==j,1,RandomMatrixEntry[i,j,pattern,maxdisturbance]]],{i,1,size},{j,1,size}];
RandomMatrix[size_, maxdisturbance_]:=RandomMatrix[size, RandomRankingPattern[size, 9], maxdisturbance];
RationalScale[scaleSize_]:= RationalScale[scaleSize]= Sort@Join[Table[p,{p,1,scaleSize}],Table[1/p,{p,2,scaleSize}]];
FundamentalScale[]:=RationalScale[9];
RandomRationalUniformRankingPattern[size_, scaleSize_]:=RandomSample[RationalScale[scaleSize],size];
(* By design this should be (scale size)^2 otherwise matrix never be consistent, even for disturbance factor 1 *)
ListOfPossibleRationalEntries[entriesLimit_]:=ListOfPossibleRationalEntries[entriesLimit] = Sort@DeleteDuplicates@Flatten@Table[i/j,{i,1,entriesLimit},{j,1,entriesLimit}];
RandomRationalMatrixEntry[i_,j_,pattern_, scaleSize_, maxdisturbance_]:=With[{randomEntry = RandomMatrixEntry[i,j,pattern,maxdisturbance], possibleEntries = ListOfPossibleRationalEntries[scaleSize]},
    With[{positionOfRepresentat = Ordering[Map[Abs[#-randomEntry] &,possibleEntries], 1]},
    First@possibleEntries[[positionOfRepresentat]]]];
RandomRationalMatrix[size_, scaleSize_, pattern_, maxdisturbance_]:=RecreatePCMatrix@ Table[If[i>j,0,If[i==j,1,RandomRationalMatrixEntry[i,j,pattern, scaleSize, maxdisturbance]]],{i,1,size},{j,1,size}];
RandomRationalMatrix[size_, maxdisturbance_]:= RandomRationalMatrix[size,9,RandomRationalUniformRankingPattern[size,9] ,maxdisturbance];

EdgeExistsQ[graph_, edgePair_]:=Length@Flatten@Position[EdgeList[graph], First@edgePair\[UndirectedEdge]Last@edgePair] > 0
RandomConnectedGraph[params_]:=If[Last@params + 1< First@params, Null, 
    With[{g = RandomGraph[params]}, If[ConnectedGraphQ@g, g, RandomConnectedGraph[params]]]];
DecompletedMatrixByGraph[M_, graph_, emptySymb_]:=MapIndexed[If[#2[[1]] == #2[[2]], 1, 
    If[EdgeExistsQ[graph, #2] || EdgeExistsQ[graph, Reverse@#2], #1, emptySymb] ]   &, M, {2}];
RandomlyDecompletedMatrix[M_, noOfMissingComparisons_, emptySymb_]:=With[{noOfcomp =((Length@M )*(Length@M)-(Length@M))/2-noOfMissingComparisons},
    With[{graph = RandomConnectedGraph[{Length@M,noOfcomp }]}, 
        DecompletedMatrixByGraph[M, graph, emptySymb]]];
        
(***************** Consistency Ratio *****************)

RationalRandomScaledMatrix[matrixSize_, scaleSize_]:=RecreatePCMatrix@Table[If[i<= j,RandomChoice[RationalScale[scaleSize]]], {i,1,matrixSize}, {j,1,matrixSize}];
RandomIndex[n_, tries_, scaleSize_]:= Mean[Map[SaatyIdx[#]&,Table[RationalRandomScaledMatrix[n,scaleSize],tries]]];
RI[n_, tries_, scaleSize_]:=RandomIndex[n, tries, scaleSize];
RandomIndex[size_]:=RandomIndex[size]=With[{cache = <|3->0.527229025896751`,4->0.8862594434407611`,5->1.1123629449050576`,6->1.2457843374586108`,7->1.340440416817517`,8->1.4036381033815553`,9->1.4536196561717807`,10->1.485308918735309`,11->1.516572262531149`,12->1.5372328925338654`|>},
If[size>2 && size <= 12, cache[size], 
If[size <= 2,-1,RandomIndex[size, 1000, 9]
]]];
RI[n_]:=RandomIndex[n];
SaatyConsistencyRatio[matrix_]:= SaatyIdx[matrix]/RandomIndex[Length@matrix];
CR[n_]:=SaatyConsistencyRatio[n];

(***************** Inconsistency indices for complete matrices *****************)

(***************** GCI - Geometric Consistency Index *****************)
GetErrorMatrix[matrix_, w_]:=MapIndexed[ #1 * w[[Last@#2]]/w[[First@#2]] &, matrix, {2}];
UpperTriangularizeToUndef[matrix_]:=MapIndexed[If[First@#2 >= Last@#2, \[Placeholder], #1] &, matrix, {2}];
(*GCI[matrix_]:=Mean@Map[(Log@#)^2 &, Select[N@Flatten@UpperTriangularizeToUndef@GetErrorMatrix[matrix, GeometricRescaledRank@matrix], NumberQ@#&]];*)
GCI[matrix_]:=2/((Length@matrix-1)(Length@matrix-2))*Plus@@Map[(Log@#)^2 &, Select[N@Flatten@UpperTriangularizeToUndef@GetErrorMatrix[matrix, GeometricRescaledRank@matrix], NumberQ@#&]];
GeometricConsistencyIndex[matrix_]:=GCI[matrix];

(***************** HCI - Harmonic Consistency Index *****************)
HM[M_,n_]:=n/Plus@@(First[{Table[1,n]} . M]^-1);
HCI[M_]:=With[{n=Length@M},((HM[M,n]-n)(n+1))/(n(n-1))];
HarmonicConsistencyIndex[M_]:=HCI[M];

(***************** CIStar - Index of Determinants *******************)
MatrixOfTriad[matrix_, triad_]:= With[{i = triad[[1]], k=triad[[2]], j=triad[[3]]}, With[{cik=matrix[[i,k]], cjk=matrix[[j,k]], cij=matrix[[i,j]]}, ({
 {1, cij, cik},
 {1/cij, 1, cjk},
 {1/cik, 1/cjk, 1}
})]];
TriadExistsQ[matrix_, triad_]:= With[{i = triad[[1]], k=triad[[2]], j=triad[[3]]}, NumberQ@matrix[[i,k]] && NumberQ@matrix[[k,j]] && NumberQ@matrix[[i,j]]];
CIStar[matrix_]:= With[{allTriads = Subsets[Range[Length@matrix], {3}]}, With[{existingTriads = Select[allTriads,TriadExistsQ[matrix, #] & ]}, Mean@Map[Det@MatrixOfTriad[matrix, #] &,  existingTriads]]];
IndexOfDeterminants[matrix_]:=CIStar[matrix];

(***************** TBII1, TBII2, TBIIAlpha, TBIIAlphaBeta - Triad based inconsistency indices *****************)

TBII1[matrix_]:= With[{allTriads = Subsets[Range[Length@matrix], {3}]}, With[{existingTriads = Select[allTriads,TriadExistsQ[matrix, #] & ]}, Mean@Map[KoczkodajTriadIndex[matrix, #] &,  existingTriads]]];
TriadBasedInconsistencyIndex1[matrix_]:=TBII1[matrix];
TBII2[matrix_]:= With[{allTriads = Subsets[Range[Length@matrix], {3}]}, With[{existingTriads = Select[allTriads,TriadExistsQ[matrix, #] & ]}, With[{trlist = Map[KoczkodajTriadIndex[matrix, #]*KoczkodajTriadIndex[matrix, #] &,  existingTriads]}, Sqrt[Plus@@trlist]/Length@trlist ]]];
TriadBasedInconsistencyIndex2[matrix_]:=TBII2[matrix];

TBIIAlpha[matrix_, alpha_]:= alpha*KoczkodajIndex@matrix +(1-alpha)*TBII1@matrix;
TriadBasedInconsistencyIndexAlpha[matrix_, alpha_]:=TBIIAlpha[matrix, alpha];

TBIIAlphaBeta[matrix_, alpha_, beta_]:= alpha*KoczkodajIndex@matrix + beta*TBII1@matrix + (1-alpha - beta)*TBII2@matrix;
TriadBasedInconsistencyIndexAlphaBeta[matrix_, alpha_, beta_]:=TBIIAlphaBeta[matrix, alpha beta];



(***************** GWI - Golden-Wang Index (for both complete and incomplete PC matrix) *****************)

CStarMatrix[matrix_]:=Transpose@Map[#/Plus@@# & ,Transpose@matrix];
CompleteUndefToZero[matrix_]:=Map[If[NumberQ@#, #, 0]&, matrix, {2}];
DeteriorateMatrixAccordingToPattern[inputMatrix_,patternMatrix_]:=Map[If[#[[2]]==0,\[Placeholder],#[[1]]] &, Transpose/@Transpose[{inputMatrix, patternMatrix}], {2}];
GMMMatrix[matrix_]:=With[{g=N@GMM@matrix}, Transpose@Table[g,{x,1,Length@matrix}]];
DiffMatrix[matrix_]:= CStarMatrix@CompleteUndefToZero@DeteriorateMatrixAccordingToPattern[GMMMatrix[matrix], CompleteUndefToZero@matrix] - CStarMatrix@CompleteUndefToZero@matrix;
PreprocessedDiffMatrix[matrix_]:= DeteriorateMatrixAccordingToPattern[Abs@DiffMatrix[matrix], CompleteUndefToZero@matrix];
(* for complete PC matrix there are 1/n^2 elements in the mean thus to get 1/n we need to multiply the expression by n *)
GWI[matrix_]:=1/Length@matrix*Plus@@Select[Flatten@PreprocessedDiffMatrix@matrix, NumberQ];
GGWI[matrix_]:=GWI[matrix];
GoldenWangIndex[matrix_]:=GWI[matrix];
GeneralizedGoldenWangIndex[matrix_]:=GWI[matrix];



(***************** SHI - Salo-Hamalainen Index *****************)

Lower[matrix_,i_,j_]:=With[{listToConsider = Select[Table[matrix[[i,k]]*matrix[[k,j]],{k,1,Length@matrix}], #!= 0 &]}, If[Length@listToConsider == 0 , 0, Min@listToConsider ]];
Upper[matrix_, i_, j_]:= With[{listToConsider = Select[Table[matrix[[i,k]]*matrix[[k,j]],{k,1,Length@matrix}], #!= 0 &]}, If[Length@listToConsider == 0 , 0, Max@listToConsider ]];
RMatrixLU[matrix_]:=Table[{Lower[matrix, i,j], Upper[matrix, i, j]}, {i, 1,Length@matrix}, {j,1,Length@matrix}];
CellExpression[r_, n_, i_, j_]:= (Last@r[[i,j]] - First@r[[i,j]])/((1+Last@r[[i,j]])(1+ First@r[[i,j]]));
SHI[matrix_]:=With[{r = RMatrixLU@CompleteUndefToZero@matrix, n=Length@matrix}, Mean@Map[CellExpression[r, n, First@#, Last@#] &, Subsets[Range[n], {2}]]];
SaloHamalainenIndex[matrix_]:=SHI[matrix];

(***************** RE - Relative Error (Barzilai's) index (for both complete and incomplete PC matrix) *****************)
(* version for multiplicative matrices *)

GetProperEntry[i_,j_, matrix_, gmm_]:=If[NumberQ@matrix[[i,j]], matrix[[i,j]], gmm[[i]]/gmm[[j]]];
GetProperEntryZero[i_,j_, matrix_, gmm_]:=If[NumberQ@matrix[[i,j]], matrix[[i,j]], 1];
OneMatrixQ[matrix_]:=And@@Map[#==1&,Flatten@matrix];
RE[matrix_]:= If[OneMatrixQ@matrix, 0, With[{gmm = GMM[matrix], n=Length@matrix}, Sum[(Log[GetProperEntry[i,j,matrix,gmm]]-Log[gmm[[i]]/gmm[[j]]])^2, {i,1,n}, {j,1,n}]/Sum[Log[GetProperEntry[i,j,matrix,gmm]]^2,{i,1,n}, {j,1,n}]]];
RelativeError[matrix_]:=RE[M];
GRE[matrix_]:=RE[M];
GeneralizedRelativeError[matrix_]:=RE[M];
RC[matrix_]:=If[OneMatrixQ@matrix, 1, With[{gmm = GMM[matrix], n=Length@matrix}, Sum[(Log[gmm[[i]]/gmm[[j]]])^2, {i,1,n}, {j,1,n}]/Sum[Log[GetProperEntry[i,j,matrix,gmm]]^2,{i,1,n}, {j,1,n}]]];
REv2[matrix_]:=If[OneMatrixQ@matrix, 0, With[{gmm = GMM[matrix], n=Length@matrix}, Sum[(Log[GetProperEntry[i,j,matrix,gmm]]-Log[gmm[[i]]/gmm[[j]]])^2, {i,1,n}, {j,1,n}]/Sum[Log[GetProperEntryZero[i,j,matrix,gmm]]^2,{i,1,n}, {j,1,n}]]];
GREv2[matrix_]:=REv2[matrix];
RelativeErrorv2[matrix_]:=REv2[M];
GeneralizedRelativeErrorv2[matrix_]:=REv2[M];

(***************** Inconsistency indices for incomplete matrices ***************)

(***************** GGCI (Generalized Geometric Consistency Index) for incomplete PC matrix *****************)
GetGMMErrorMatrix[matrix_]:=Map[If[NumberQ@#, #, \[Placeholder]] &, GetErrorMatrix[matrix, GMM@matrix], {2}]; 
GGCI1[matrix_]:=Mean@Map[(Log@#)^2 &, Select[Flatten@UpperTriangularizeToUndef@GetGMMErrorMatrix@matrix, NumberQ@#&]];
GeneralizedGeometricConsistencyIndex1[matrix_]:=GGCI1[matrix];
GGCI2[matrix_]:=With[{n=Length@matrix},2/((n-1)(n-2)) Plus@@Map[(Log@#)^2 &, Select[Flatten@UpperTriangularizeToUndef@GetGMMErrorMatrix@matrix, NumberQ@#&]]];
GeneralizedGeometricConsistencyIndex2[matrix_]:=GGCI2[matrix];

(***************** GKI (Generalized Koczkodaj Index) for incomplete PC matrix *****************)

KoczkodajTupleIndex[matrix_, tuple_]:=Min[Abs[1-Product[matrix[[tuple[[i]],tuple[[i+1]]]], {i, 1, Length@tuple-1}]/matrix[[tuple[[1]],  tuple[[-1]]]]], Abs[1-matrix[[tuple[[1]],  tuple[[-1]]]]/Product[matrix[[tuple[[i]],tuple[[i+1]]]], {i, 1, Length@tuple-1}]]];
TupleExistsQ[matrix_, tuple_]:=And@@Table[NumberQ@matrix[[tuple[[i]], tuple[[i+1]]]], {i,1,Length@tuple-1}] && NumberQ@matrix[[tuple[[1]], tuple[[-1]]]];
GetTuplesByCyclesFromAdjacencyMatrix[adjMatrix_]:=GetTuplesByCyclesFromAdjacencyMatrix[adjMatrix]=Map[First@# &, FindCycle[AdjacencyGraph@adjMatrix, Infinity, All], {2}];
AdjacencyMatrixFromPCMatrix[matrix_]:=Map[If[NumberQ@#, 1, 0] &, matrix, {2}]-IdentityMatrix[Length@matrix];
GetTuplesByCycles[matrix_]:=GetTuplesByCyclesFromAdjacencyMatrix@AdjacencyMatrixFromPCMatrix@matrix;
GKI[matrix_]:=  With[{tuples = GetTuplesByCycles[matrix] }, If[Length@tuples == 0, 0, Max@Map[KoczkodajTupleIndex[matrix, #] &,  tuples]]];
GeneralizedKoczkodajIndex[matrix_]:=GKI[matrix];

(***************** GTBII1, GTBII2, GTBIIAlpha, GTBIIAlphaBeta - Generalized tuple based inconsistency indices for incomplete PC matrices *****************)

GTBII1[matrix_]:= With[{tuples = GetTuplesByCycles[matrix] }, If[Length@tuples == 0, 0,Mean@Map[KoczkodajTupleIndex[matrix, #] &,  tuples]]];
GeneralizedTupleBasedInconsistencyIndex1[matrix_]:=GTBII1[matrix];

GTBII2[matrix_]:=With[{tuples = GetTuplesByCycles[matrix] },If[Length@tuples == 0, 0,With[{trlist = Map[(KoczkodajTupleIndex[matrix, #])^2 &,   tuples]},Sqrt[Plus@@trlist]/Length@trlist ]]];
GeneralizedTupleBasedInconsistencyIndex2[matrix_]:=GTBII2[matrix];

GTBIIAlpha[matrix_, alpha_]:= alpha*GeneralizedKoczkodajIndex@matrix +(1-alpha)*GTBII1@matrix;
GeneralizedTupleBasedInconsistencyIndexAlpha[matrix_, alpha_]:=GTBIIAlpha[matrix, alpha];

GTBIIAlphaBeta[matrix_, alpha_, beta_]:= alpha*GKI@matrix + beta*GTBII1@matrix + (1- alpha - beta)*GTBII2@matrix;
GeneralizedTupleBasedInconsistencyIndexAlphaBeta[matrix_, alpha_, beta_]:=GTBIIAlphaBeta[matrix, alpha, beta];

(***************** GSHI (Generalized Salo-Hamalainen index) for incomplete PC matrix *****************)

GetTuplesByPath[adjmatrix_, i_,j_]:=GetTuplesByPath[adjmatrix, i,j]=FindPath[AdjacencyGraph@adjmatrix, i, j, Infinity, All];
GetProductByTuple[matrix_, tuple_]:= Times@@Table[matrix[[tuple[[i]], tuple[[i+1]]]], {i,1,Length@tuple-1}];
GetEstimationList[matrix_,i_,j_]:=Map[GetProductByTuple[matrix, #] &, GetTuplesByPath[AdjacencyMatrixFromPCMatrix@matrix, i,j]];
RMatrix[matrix_]:=Table[With[{estList = GetEstimationList[matrix, i, j]}, {Min@estList, Max@estList}], {i, 1,Length@matrix}, {j,1,Length@matrix}];
GSHI[matrix_]:=With[{n=Length@matrix, rm=RMatrix[matrix]},2/(n(n-1)) Sum[Sum[(Last@rm[[i,j]]-First@rm[[i,j]])/((1+Last@rm[[i,j]])(1+First@rm[[i,j]])),{j,i+1,n}], {i,1,n-1}]];
GeneralizedSaloHamalainenIndex[matrix_]:=GSHI[matrix];

(***************** PC and Graph representation TODO *****************)

UpperTriangularizePCM[M_]:= MapIndexed[If[#2[[1]] <= #2[[2]] ,#1, \[Infinity]]&, M , {2}];
LowerTriangularizePCM[M_]:= MapIndexed[If[#2[[1]] >= #2[[2]] ,#1, \[Infinity]]&, M , {2}];
GraphOfPCM[matrix_]:=WeightedAdjacencyGraph[MapIndexed[If[NumberQ@#1,If[#2[[1]]!= #2[[2]],#1,\[Infinity]], \[Infinity]] &, matrix, {2}]];
GraphOfGeneralizedPCM[matrix_]:=AdjacencyGraph[MapIndexed[If[NumberQ@#1  || ListQ@#1,If[#2[[1]]!= #2[[2]],1,0], 0] &, matrix, {2}]];
PlotPCGraph[M_,direction_, layout_, edgeLabelFormatter_, vertexLabelFormatter_, vertexSize_, edgeLabelSize_, vertexLabelSize_]:=With[{alpha = 1.25, cmg = If[direction, GraphOfPCM@M, UndirectedGraph@GraphOfPCM@M]}, 
    GraphPlot[Graph[cmg, EdgeLabels->Map[#-> edgeLabelFormatter@M[[#[[1]],#[[2]]]] &,EdgeList@cmg ], 
        VertexLabels->Table[i->Placed[vertexLabelFormatter@i,{1/2,1/2}],{i,Length@VertexList@cmg}], GraphLayout->layout],
        EdgeLabelStyle->Directive[Black,Plain,edgeLabelSize], VertexSize->vertexSize, VertexLabelStyle->Directive[Black,Plain,vertexLabelSize], VertexStyle->RGBColor[alpha*0.66,alpha*0.71,alpha*0.79,1]]];
PlotPCGraph[M_]:= PlotPCGraph[M,False, "CircularEmbedding", NicePlotFraction, Function[x,x], 0.15, 13, 15];
WeightListByEdgeList[wam_, edgeList_]:=Map[First@# \[UndirectedEdge] Last@# -> wam[[First@#, Last@#]]&, edgeList];
FindAllSpanningTrees[graph_]:= Module[{wam = Normal@WeightedAdjacencyMatrix@graph},
    Select[Map[Graph[#, EdgeWeight-> WeightListByEdgeList[wam, #]] &, Subsets[EdgeList@graph, {VertexCount@graph - 1}]], TreeGraphQ@# &]];
PlotUndirectedWeightedGraph[graph_, matrix_, layout_, edgeLabelFormatter_, vertexLabelFormatter_, vertexSize_, edgeLabelSize_, vertexLabelSize_]:=With[{alpha = 1.25}, 
    GraphPlot[Graph[graph, EdgeLabels->Map[#-> edgeLabelFormatter@matrix[[#[[1]],#[[2]]]] &,EdgeList@graph ],
        VertexLabels->Table[i->Placed[vertexLabelFormatter@i,{1/2,1/2}],{i,Length@VertexList@graph}], GraphLayout->layout],
        EdgeLabelStyle->Directive[Black,Plain,edgeLabelSize], VertexSize->vertexSize, VertexLabelStyle->Directive[Black,Plain,vertexLabelSize], VertexStyle->RGBColor[alpha*0.66,alpha*0.71,alpha*0.79,1]]];
PlotUndirectedWeightedGraph[graph_, matrix_]:= PlotUndirectedWeightedGraph[graph, matrix, Automatic,  NicePlotFraction, Function[x,x], 0.15, 13, 15];
(* RankBySpanningTree[sptree_, matrix_]:=Module[{rank, root = 1},
   rank = Table[\[Infinity], Length[matrix]];
   rank[[root]] = 1;
   DepthFirstScan[sptree,1,{"FrontierEdge"-> ((If[NumberQ@rank[[#[[1]]]], rank[[#[[2]]]]=  rank[[#[[1]]]]* matrix[[#[[2]], #[[1]]]], rank[[#[[1]]]]=  1/rank[[#[[2]]]]* matrix[[#[[2]], #[[1]]]]]) &)}]; Return[rank]]; *)
(* there is an assumption that all the vertices can be achieved from vertex number 1 *)
RankBySpanningTree[sptree_, M_]:=Module[{path={}, weights={1}}, 
	DepthFirstScan[sptree,1,{"PrevisitVertex"->(AppendTo[path, #]&)}];
	For[i=1, i< Length@path, i++, AppendTo[weights, weights[[i]]*M[[path[[i+1]],path[[i]]]]]];
	Return@Map[Last@# &, SortBy[Transpose[{path, weights}], First]];
];  
RankAndPathBySpanningTree[sptree_, M_]:=Module[{path={}, weights={1}}, 
    DepthFirstScan[sptree,1,{"PrevisitVertex"->(AppendTo[path, #]&)}];
    For[i=1, i< Length@path, i++, AppendTo[weights, weights[[i]]*M[[path[[i+1]],path[[i]]]]]];
    Return@Transpose[{path, weights}];
];
GetRandomSpanningTreeFromPCMatrix[M_]:=Module[{RandomMPattern, g, sp, mx, mxCorr},
(* generate random patter with accordance with M *)
RandomMPattern=Table[If[i==j || NumberQ[M[[i,j]]]==False, Infinity, Random[]], {i,1,Length@M},{j,1,Length@M}]; 
(* create graph with random weigths *)
g = WeightedAdjacencyGraph@RandomMPattern;
(* find minimal spanning tree with random vertex based on graph with random weights *)
sp = FindSpanningTree[{g,RandomInteger[{1,Length@M}]}];
(* get adjacency matrix with random weights *)
mx = Normal@WeightedAdjacencyMatrix@sp;
(* replace random weights by those from input M *)
mxCorr = Table[If[mx[[i,j]]!=0,  M[[i,j]], 
If[mx[[j,i]]!=0, 1/ M[[j,i]], Infinity]]
, {i,1,Length@VertexList@sp}, {j, 1, Length@VertexList@sp}];
(* convert matrix to graph / tree *)
Return[WeightedAdjacencyGraph@mxCorr];
];


(***********************************************************************************************************)


End[]
EndPackage[]




